/**
 * aTalk, the OpenSource Java VoIP and Instant Messaging client.
 * <p>
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.jivesoftware.smackx.avatar.vcardavatar;

import android.text.TextUtils;

import org.jivesoftware.smack.*;
import org.jivesoftware.smack.SmackException.*;
import org.jivesoftware.smack.XMPPException.XMPPErrorException;
import org.jivesoftware.smack.filter.*;
import org.jivesoftware.smack.packet.*;
import org.jivesoftware.smack.roster.*;
import org.jivesoftware.smack.util.StringUtils;
import org.jivesoftware.smackx.avatar.cache.*;
import org.jivesoftware.smackx.avatar.vcardavatar.listener.VCardAvatarListener;
import org.jivesoftware.smackx.avatar.vcardavatar.packet.VCardTempXUpdateExtension;
import org.jivesoftware.smackx.vcardtemp.VCardManager;
import org.jivesoftware.smackx.vcardtemp.packet.VCard;
import org.jxmpp.jid.*;
import org.jxmpp.jid.impl.JidCreate;
import org.jxmpp.stringprep.XmppStringprepException;
import org.jxmpp.util.cache.LruCache;

import java.io.File;
import java.security.*;
import java.util.*;
import java.util.logging.*;

/**
 * This class deals with the avatar data in XEP-0054: vcard-temp
 * The class provides a persistent storage in memoryCache and filesystem if specified.
 * Basing on XEP-0153 vCard-Based Avatars protocol, this class can be configured to auto
 * retrieve the avatar from server and store in the persistent storage for later retrieval.
 *
 * @author Eng Chong Meng
 */
public class VCardAvatarManager extends Manager
		implements StanzaListener
{
	/**
	 * The logger.
	 */
	private static final Logger LOGGER = Logger.getLogger(VCardAvatarManager.class.getName());

	/**
	 * This presence x-extension element name.
	 */
	private static final String ELEMENT = VCardTempXUpdateExtension.ELEMENT;

	/**
	 * This presence x-extension namespace.
	 */
	private static final String NAMESPACE = VCardTempXUpdateExtension.NAMESPACE;

	private static Map<XMPPConnection, VCardAvatarManager> instances = new WeakHashMap<>();

	/**
	 * Map of avatarHash" to Avatar byte data
	 */
	private static final LruCache<String, byte[]> cacheAvatar = new LruCache<>(1000);

	/**
	 * Use for the persistent avatar storage in additional to cacheAvatar
	 */
	private static AvatarCacheFile persistentAvatarCache = null;

	/**
	 * Map of bareJid to avatarId (Hash of avatar). Definition of avatarId:
	 * 1. null ==> not allow (no defined)
	 * 2. "" ==> user has no photo specified
	 * 3. {avatar Hash} ==> photo in cache and/or persistent storage
	 *
	 * Note: Server stores only one copy of VCard Info (avatar) for each jid irrespective of its
	 * resources.
	 */
	private static final LruCache<BareJid, String> cacheJidToAvatarId = new LruCache<>(1000);

	/**
	 * Use for the persistent JidToHash Index storage in additional to cacheJidToAvatarId
	 */
	private static JidToHashCacheFile persistentJidToHashIndex = null;

	/**
	 * Listeners to be informed if there is new avatar updated.
	 */
	private final List<VCardAvatarListener> mListeners = new LinkedList<>();

	/**
	 * The VCardTempXUpdate Extension class.
	 */
	private final VCardTempXUpdateExtension vCardTempXUpdateExtension;

	/**
	 * The VCard info for this user.
	 */
	private VCard mVCard = null;

	/**
	 * If <tt>true</tt>, then proceed to download VCard when a new photoHash is received from
	 * <presence/>. Default to false as client application may want to extract VCard info other
	 * than the photo image
	 */
	private static boolean mAutoDownload = false;

	/**
	 * The VCardManager associated with the VCardAvatarManager.
	 */
	private final VCardManager vCardMgr;

	/**
	 * The registered account (bareJid) for this connection.
	 */
	private BareJid mAccount;
	private XMPPConnection mConnection;

	/**
	 * Creates a filter to only listen to presence stanza with the element name "x" and the
	 * namespace "vcard-temp:x:update".
	 */
	private static final StanzaFilter PRESENCES_WITH_VCARD
			= new AndFilter(new StanzaTypeFilter(Presence.class),
			new StanzaExtensionFilter(ELEMENT, NAMESPACE));

	private static final StanzaFilter PRESENCES_WITHOUT_VCARD
			= new AndFilter(new
			StanzaTypeFilter(Presence.class),
			new NotFilter(new StanzaExtensionFilter(ELEMENT, NAMESPACE)));

	static {
		XMPPConnectionRegistry.addConnectionCreationListener(new ConnectionCreationListener()
		{
			public void connectionCreated(XMPPConnection connection)
			{
				getInstanceFor(connection);
			}
		});
	}

	public static synchronized VCardAvatarManager getInstanceFor(XMPPConnection connection)
	{
		VCardAvatarManager vCardAvatarManager = instances.get(connection);
		if (vCardAvatarManager == null) {
			vCardAvatarManager = new VCardAvatarManager(connection);
		}
		return vCardAvatarManager;
	}

	/**
	 * Create an UserAvatarManager.
	 *
	 * @param connection
	 * 		the registered account XMPP connection
	 */
	private VCardAvatarManager(XMPPConnection connection)
	{
		super(connection);
		vCardMgr = VCardManager.getInstanceFor(connection);
		vCardTempXUpdateExtension = new VCardTempXUpdateExtension(null);
		instances.put(connection, this);

		connection.addConnectionListener(new AbstractConnectionListener()
		{
			@Override
			public void connected(XMPPConnection connection)
			{
				mConnection = connection;
			}

			/**
			 * Upon user authentication, update account avatarHash so it is ready for
			 * x-extension inclusion in <presence/> sending.
			 * @see XMPPConnection#addPacketInterceptor(StanzaListener, StanzaFilter)
			 * @see VCardAvatarManager#processPacket(Stanza)
			 *
			 * Application can also update the avatarHash anytime using
			 * @see VCardTempXUpdateExtension#setAvatarHash(String)
			 */
			@Override
			public void authenticated(XMPPConnection connection, boolean resumed)
			{
				mAccount = connection.getUser().asBareJid();
				if (persistentJidToHashIndex != null) {
					String hash = getAvatarHashByJid(mAccount);
					if (hash != null)
						vCardTempXUpdateExtension.setAvatarHash(hash);
				}
			}
		});

		/** Intercepts all sent <presence/> stanza in order to add the photo tag. */
		connection.addPacketInterceptor(this, new StanzaTypeFilter(Presence.class));

		/**
		 * Listen for remote presence stanzas with the vCardTemp:x:update extension. If we
		 * receive such a stanza, process the stanza and acts if necessary
		 */
		connection.addAsyncStanzaListener(new StanzaListener()
		{
			@Override
			public void processPacket(Stanza stanza)
			{
				parseContactPhotoPresence(stanza);
			}

		}, PRESENCES_WITH_VCARD);

		/** XEP-0135 protocol not supported if x-extension no found in <presence/> */
		connection.addAsyncStanzaListener(new StanzaListener()
		{
			@Override
			public void processPacket(Stanza packet)
			{
				LOGGER.log(Level.INFO, "This account does not support XEP-0135: "
						+ packet.getFrom());
			}
		}, PRESENCES_WITHOUT_VCARD);
	}

	/**
	 * Set to false (DEFAULT) if application needs to extract all VCard info other than
	 * the photo image
	 *
	 * @param autoDownload
	 * 		true to enable auto download of avatars (photo image)
	 */
	public static void setAutoDownload(boolean autoDownload)
	{
		mAutoDownload = autoDownload;
	}

	/**
	 * Set the persistent cache implementation
	 *
	 * @param storeDir
	 * 		the file directory which will store the avatars
	 */
	public static void setPersistentCache(File storeDir)
	{
		if (storeDir != null) {
			persistentJidToHashIndex = new JidToHashCacheFile(storeDir);
			persistentAvatarCache = new AvatarCacheFile(storeDir);
		}
	}

	/**
	 * If avatar is new, add it to the cache; and persistent storage database if enabled.
	 *
	 * @param avatarId
	 * 		The avatarId, i.e SHA-1 Hash of the avatarImage.
	 * @param avatarImage
	 * 		Byte[] value of avatar image data.
	 */
	public static void addAvatarImageByAvatarId(String avatarId, byte[] avatarImage)
	{
		if (!TextUtils.isEmpty(avatarId) && isAvatarNew(avatarId)) {
			cacheAvatar.put(avatarId, avatarImage);
			if (persistentAvatarCache != null)
				persistentAvatarCache.addIAvatarByHash(avatarId, avatarImage);
		}
	}

	/**
	 * Calculate the avatarId and save its image to cache
	 *
	 * @param avatarImage
	 * 		Byte[] value of avatar image data.
	 * @return <tt>avatarId</tt> i.e SHA-1 Hash of the avatarImage.
	 */
	public static String addAvatarImage(byte[] avatarImage)
	{
		if (avatarImage.length == 0)
			return "";

		String avatarId = getAvatarHash(avatarImage);
		if (!TextUtils.isEmpty(avatarId)) {
			addAvatarImageByAvatarId(avatarId, avatarImage);
		}
		return avatarId;
	}

	/**
	 * Get an avatar image from the cache with the given avatarId
	 *
	 * @param avatarId
	 * 		the id of the avatar (Hash)
	 * @return the avatar or null if it cannot be retrieved from the cache or file
	 */
	public static byte[] getAvatarImageByHash(String avatarId)
	{
		if (avatarId == null)
			return null;

		byte[] avatarImage = new byte[0];  // default to no photo
		if (avatarId.length() != 0) {
			avatarImage = cacheAvatar.get(avatarId);
			if ((avatarImage == null) && (persistentAvatarCache != null)) {
				avatarImage = persistentAvatarCache.getAvatarForHash(avatarId);
				if (avatarImage != null) {
					cacheAvatar.put(avatarId, avatarImage);
				}
			}
		}
		return avatarImage;
	}

	/**
	 * Get the Avatar for a jid. Returns the avatar or null if
	 * UserAvatarManager does not have any information.
	 *
	 * @param jid
	 * 		the user (BareJid)
	 * @return the avatarImage or null
	 */
	public static byte[] getAvatarImageByJid(BareJid jid)
	{
		String avatarId = getAvatarHashByJid(jid);
		if (avatarId == null)
			return null;

		LOGGER.log(Level.INFO, "Fetching VCard avatar from cache for: (" + jid + ") => "
				+ avatarId);
		return getAvatarImageByHash(avatarId);
	}

	/**
	 * Get the avatarHash for jid
	 *
	 * @param jid
	 * 		the bareJid
	 * @return the avatarHash for the jid or null
	 */
	public static String getAvatarHashByJid(BareJid jid)
	{
		String hash = cacheJidToAvatarId.get(jid);
		if ((hash == null) && (persistentJidToHashIndex != null)) {
			hash = persistentJidToHashIndex.getHashForJid(jid);
			if (hash != null) {
				cacheJidToAvatarId.put(jid, hash);
			}
		}
		return hash;
	}

	/**
	 * add a new JidToAvatar Index to the cache
	 *
	 * @param userId
	 * 		the bareJid
	 * @param avatarHash
	 * 		the avatar Hash cannot be empty
	 */
	private static void addJidToAvatarHashIndex(BareJid userId, String avatarHash)
	{
		// Create an index hash for the jid
		if (!TextUtils.isEmpty(avatarHash)) {
			cacheJidToAvatarId.put(userId, avatarHash);
			if (persistentJidToHashIndex != null) {
				persistentJidToHashIndex.addHashByJid(userId, avatarHash);
			}
		}
	}

	/**
	 * Check if we have the user avatar in store
	 *
	 * @param avatarId
	 * 		the id of the avatar (Hash)
	 * @return <tt>true</tt> if avatarId is avatarId is new. <tt>false</tt> otherwise
	 */
	private static boolean isAvatarNew(String avatarId)
	{
		boolean isFound = cacheAvatar.containsKey(avatarId);
		if (!isFound && (persistentAvatarCache != null)) {
			isFound = persistentAvatarCache.contains(avatarId);
		}
		return !isFound;
	}

	/**
	 * Check if the given avatarHash is owned by multiple user.
	 *
	 * @param userId
	 * 		the known owner of the avatarHash
	 * @param avatarHash
	 * 		the id of the avatar (Hash)
	 * @return <tt>true</tt> if avatarHash is owned by more than one user. <tt>false</tt> otherwise
	 */
	private boolean isHashMultipleOwner(BareJid userId, String avatarHash)
	{
		boolean isMultipleUser = false;

		Roster roster = Roster.getInstanceFor(mConnection);
		Set<RosterEntry> rosterEntries = roster.getEntries();
		for (RosterEntry rosterEntry : rosterEntries) {
			BareJid contactJid = rosterEntry.getJid();
			if (!contactJid.equals(userId.toString())) {
				String imageHash = getAvatarHashByJid(contactJid);
				if (!TextUtils.isEmpty(imageHash) && imageHash.equals(avatarHash)) {
					isMultipleUser = true;
					break;
				}
			}
		}
		return isMultipleUser;
	}

	/**
	 * Save a copy of the account Roster's contact names in persistent storage delimited by ",".
	 * Later used by method clearPersistentStorage() when account is removed even when offline.
	 *
	 * rosterFileName is mimic as BareJid so as to use common persistent storage
	 *
	 * @param account
	 * 		the roster information belong to this account
	 * @see #clearPersistentStorage(BareJid)
	 */
	public void saveAccountRoster(BareJid account)
			throws XmppStringprepException
	{
		if ((account != null) && (persistentJidToHashIndex != null)) {
			BareJid rosterFileName = JidCreate.bareFrom(account + "_roster");

			StringBuilder sb = new StringBuilder();
			Roster roster = Roster.getInstanceFor(mConnection);
			Set<RosterEntry> rosterEntries = roster.getEntries();

			for (RosterEntry rosterEntry : rosterEntries) {
				String contact = rosterEntry.getJid() + ",";
				sb.append(contact);
			}

			sb.deleteCharAt(sb.length() - 1);
			persistentJidToHashIndex.addHashByJid(rosterFileName, sb.toString());
		}
	}

	/**
	 * Clear the persistent storage for the given account even when is offline:
	 * - User has unregistered the account
	 * - Allow user to only clean avatar info pertaining to the specific account
	 * - User roster group change
	 *
	 * @param account
	 * 		all the avatar information belong to this account are to be purged
	 */
	public static void clearPersistentStorage(BareJid account)
			throws XmppStringprepException
	{
		if ((account != null) && (persistentJidToHashIndex != null)) {

			BareJid rosterFileName = JidCreate.bareFrom(account + "_roster");
			String rosterContacts = persistentJidToHashIndex.getHashForJid(rosterFileName);

			// Remove avatar info for all contacts of this account
			if (rosterContacts != null) {
				String[] contacts = rosterContacts.split(",");
				for (String contact : contacts) {
					BareJid contactJid = JidCreate.bareFrom(contact);

					String imageHash = getAvatarHashByJid(contactJid);
					if (!TextUtils.isEmpty(imageHash)) {
						persistentJidToHashIndex.purgeItemFor(contactJid);
						if (persistentAvatarCache != null)
							persistentAvatarCache.purgeItemFor(imageHash);
					}
				}
				// Remove the account roster info file
				persistentJidToHashIndex.purgeItemFor(rosterFileName);

				// Finally remove account own avatar info
				String imageHash = getAvatarHashByJid(account);
				if (!TextUtils.isEmpty(imageHash))
					persistentAvatarCache.purgeItemFor(imageHash);
				persistentJidToHashIndex.purgeItemFor(account);
			}
		}
	}

	/**
	 * Clear all the VCard persistent storage for the device when:
	 * - to cleanup and refresh VCard fetching
	 * - Persistent storage option has been disabled by user
	 */
	public static void clearPersistentStorage()
	{
		if (persistentJidToHashIndex != null)
			persistentJidToHashIndex.emptyCache();
		if (persistentAvatarCache != null)
			persistentAvatarCache.emptyCache();
	}

	/**
	 * Download VCard for the given userId, and update avatar information if available.
	 * A null userId indicates that the download is for the connection registered account.
	 * <p>
	 * Do not cache avatarId in persistent store if VCard does not have <PHOTO/>#photoBinval item.
	 * We may never get the updated VCard info if contact client does not support XEP-0153 protocol
	 *
	 * @param userId
	 * 		The bareJid of the user. Load VCard of the current user if null.
	 * @return VCard if the download was successful otherwise null is returned
	 */
	public VCard downloadVCard(BareJid userId)
	{
		if (userId == null)
			userId = mAccount;

		try {
			mVCard = vCardMgr.loadVCard((EntityBareJid) userId);
		}
		catch (SmackException.NoResponseException | XMPPException.XMPPErrorException
				| SmackException.NotConnectedException | InterruptedException e) {
			LOGGER.log(Level.SEVERE, "Error while downloading VCard for: " + userId);
		}

		if (mVCard != null) {
			String currentAvatarHash = getAvatarHashByJid(userId);

			String avatarHash = "";  // default to no photo specified
			byte[] avatarImage = mVCard.getAvatar();
			// Proceed only if vCard has photo specified
			if (avatarImage != null) {
				// save if new image to persistent cache and get its avatarHash
				avatarHash = addAvatarImage(avatarImage);
			}

			if (!avatarHash.equals(currentAvatarHash)) {
				/**
				 * set <presence/> avatarHash if the userId is the registered account for this
				 * connection
				 * @see VCardTempXUpdateExtension#mAvatarHash
				 */
				if (mAccount.equals(userId)) {
					vCardTempXUpdateExtension.setAvatarHash(avatarHash);
				}

				// add an index hash for the jid
				addJidToAvatarHashIndex(userId, avatarHash);

				// Purge old item if the currentAvatarHash is single user owner
				if (!TextUtils.isEmpty(currentAvatarHash)
						&& !isHashMultipleOwner(userId, currentAvatarHash))
					persistentAvatarCache.purgeItemFor(currentAvatarHash);
			}
			LOGGER.log(Level.INFO, "Downloaded vcard info for: " + userId + "; Hash = "
					+ avatarHash);
		}
		return mVCard;
	}

	/**
	 * Save this vCard for the user connected by 'connection'. XMPPConnection should be
	 * authenticated and not anonymous.
	 *
	 * @throws XMPPErrorException
	 * 		thrown if there was an issue setting the VCard in the server.
	 * @throws NoResponseException
	 * 		if there was no response from the server.
	 * @throws NotConnectedException
	 */

	public boolean saveVCard(VCard vCard)
			throws NoResponseException, XMPPErrorException, NotConnectedException,
			InterruptedException
	{
		boolean isImageUpdated = false;
		if (vCard != null) {
			isImageUpdated = updateVCardAvatarHash(vCard.getAvatar());
			vCardMgr.saveVCard(vCard);
		}
		return isImageUpdated;
	}

	/**
	 * Gets the String representation in hexadecimal of the SHA-1 hash of the image given in
	 * parameter.
	 *
	 * @param imageData
	 * 		The image to getHashForJid the hexadecimal representation of the SHA-1 hash.
	 * @return The SHA-1 hash hexadecimal representation of the image. Null if the image is null or
	 * if the SHA1 is not recognized as a valid algorithm.
	 */
	public static String getAvatarHash(byte[] imageData)
	{
		byte[] imageHash = null;
		String avatarHash = null;

		try {
			if (imageData != null) {
				MessageDigest messageDigest = MessageDigest.getInstance("SHA1");
				imageHash = messageDigest.digest(imageData);
			}
		}
		catch (NoSuchAlgorithmException ex) {
			LOGGER.log(Level.SEVERE, "Failed to getHashForJid message digest: " + ex);
		}

		if (imageHash != null)
			avatarHash = StringUtils.encodeHex(imageHash);

		return avatarHash;
	}

	/**
	 * Updates the avatar image manged by this presence extension. Should only be called from
	 * registered account of this connection.
	 *
	 * @param imageBytes
	 * 		The avatar image and must not be null
	 * @return "false" if the new avatar image is the same as the current one. "true" if this
	 * presence extension has been updated with the new avatar image.
	 */
	public boolean updateVCardAvatarHash(byte[] imageBytes)
	{
		if (imageBytes == null)
			return false;

		boolean isImageUpdated = false;
		if (imageBytes.length != 0) {
			// Computes the SHA-1 hash in hexadecimal representation.
			String avatarHash = getAvatarHash(imageBytes);

			if (isAvatarNew(avatarHash)) {
				vCardTempXUpdateExtension.setAvatarHash(avatarHash);

				// If the image has changed, then update all caches info and cache image content
				addAvatarImageByAvatarId(avatarHash, imageBytes);

				// Note: user does not have photo if avatar == ""
				addJidToAvatarHashIndex(mAccount, avatarHash);

				LOGGER.log(Level.INFO, "Avatar Hash updated for : " + mAccount + "; Hash = "
						+ avatarHash);
				isImageUpdated = true;
			}
		}
		return isImageUpdated;
	}

	/* ===================================================================================== */

	/**
	 * Parses a contact presence stanza with the element name "x" and the namespace
	 * "vcard-temp:x:update". If new avatarHash received, Download if autoDownload is
	 * enabled and/or fireListeners registered with this event if change detected.
	 * <p>
	 *
	 * @param stanza
	 * 		The stanza received to parse.
	 */
	private void parseContactPhotoPresence(Stanza stanza)
	{
		// Retrieves the user bareJid and its current avatarHash
		Jid userId = stanza.getFrom();
		String currentAvatarHash = getAvatarHashByJid(userId.asBareJid());

		// Get the stanza extension which contains the photo tag.
		VCardTempXUpdateExtension vCardXExtension = stanza.getExtension(ELEMENT, NAMESPACE);

		/**
		 * Retrieved photoHash info may contains [null | "" | "{avatarHash}"
		 *
		 */
		if (vCardXExtension != null) {
			String avatarHash = vCardXExtension.getAvatarHash();

			/* acts if only new avatarHash is received. null => client not ready so no action*/
			if ((avatarHash != null) && isAvatarNew(avatarHash)) {
				/**
				 * If autoDownload is enabled, download VCard and it will also update all
				 * relevant avatar information if download is successful
				 */
				if (mAutoDownload) {
					mVCard = downloadVCard(userId.asBareJid());
				}
				else {
					// Invalid mVcard on new avatarHash received
					mVCard = null;
				}

				LOGGER.log(Level.INFO, "Presence with new avatarHash received (old => new) from: "
						+ userId + "\n" + currentAvatarHash + "\n" + avatarHash);
				fireListeners(userId, avatarHash);
			}
		}
	}

	/* ===================================================================================== */

	/**
	 * Intercepts sent presence packets in order to add VCardTempXUpdateExtension extension.
	 *
	 * @param stanza
	 * 		The sent presence packet.
	 */
	@Override
	public void processPacket(Stanza stanza)
			throws SmackException.NotConnectedException
	{
		LOGGER.log(Level.INFO, "Presence x-extension added to sent stanza for account: "
				+ mConnection.getUser() + " <= " + stanza.toXML());
		stanza.addExtension(vCardTempXUpdateExtension);
	}

	/* ===================================================================================== */

	/**
	 * Fire the listeners if there is a change in the avatarHash, and the auto downloaded VCard
	 * info if any; otherwise null vCard is sent to the listeners
	 *
	 * @param from
	 * 		the full jid of the contact sending the <presence/> stanza
	 * @param avatarId
	 * 		the new avatar id can be "" or {avatarHash}
	 */
	private void fireListeners(Jid from, String avatarId)
	{
		for (VCardAvatarListener l : mListeners)
			l.onAvatarChange(from, avatarId, mVCard);
	}

	/**
	 * Add an VCardAvatarListener that will be informed if there is change in the avatarHash.
	 *
	 * @param listener
	 * 		the VCardAvatarListener to add
	 */
	public void addVCardAvatarChangeListener(VCardAvatarListener listener)
	{
		if (!mListeners.contains(listener))
			mListeners.add(listener);
	}

	/**
	 * Remove an VCardAvatarListener.
	 *
	 * @param listener
	 * 		the VCardAvatarListener to remove
	 */
	public void removeVCardAvatarChangeListener(VCardAvatarListener listener)
	{
		mListeners.remove(listener);
	}
}
